// llmongo.go
package main

import (
	"flag"
	"fmt"
	"github.com/fzzy/radix/redis"
	"gopkg.in/mgo.v2"
	"os"
	//    "gopkg.in/mgo.v2/bson"
	"log"
	"path"
	"path/filepath"
	"strconv"
	"strings"
	"time"
)

type Options struct {
	MongoUri string
	RedisUri string
	rdb      *redis.Client
	mdb      *mgo.Session
	ll       *mgo.Collection
	RedisTTL time.Duration
	LogFile  string
	Version  bool
}

type MongoLogin struct {
	User     string    `json:"user"`
	Protocol string    `json:"protocol"`
	Ip       string    `json:"ip"`
	Date     time.Time `json:"date"`
}

type Index struct {
	User string    `json:"user"`
	Date time.Time `json:"date"`
}

const (
	_VERSION = "v1.1.0"
)

var (
	opts = Options{
		MongoUri: "mongodb://127.0.0.1:27018",
		RedisUri: "redis-ll.mail.tiscali.sys:6379",
		//MongoUri: "mongodb://10.39.81.85:27018",
		//RedisUri: "127.0.0.1:6379",
		RedisTTL: time.Hour * 11688, // 16 mesi
		LogFile:  "log/llmongo.log",
	}
)

func usage() {
	fmt.Println("Usage: llmongo -m <mongo uri> -r <redis uri> -t <ttl> -l <logfile> -v\n")
	os.Exit(0)
}

func init() {
	current, err := filepath.Abs(filepath.Dir(os.Args[0]))
	if err != nil {
		log.Fatal(err)
	}

	opts.LogFile = path.Join(current, opts.LogFile)

	//	fmt.Println(opts.LogFile)

	flag.StringVar(&opts.MongoUri, "m", opts.MongoUri, "Mongodb")
	flag.StringVar(&opts.RedisUri, "r", opts.RedisUri, "Redis")
	flag.StringVar(&opts.LogFile, "l", opts.LogFile, "Logs filename")
	flag.DurationVar(&opts.RedisTTL, "t", opts.RedisTTL, "Redis TTL")
	flag.BoolVar(&opts.Version, "v", false, "Version")
}

func connectRedis() {
	var err error
	opts.rdb, err = redis.Dial("tcp", opts.RedisUri)
	if err != nil {
		log.Println("Redis connect Error: ", err.Error())
		os.Exit(-1)
	}
}

func connectMongo() {
	var err error
	opts.mdb, err = mgo.Dial(opts.MongoUri)
	if err != nil {
		log.Println("Mongodb connect Error: ", err.Error())
		os.Exit(-3)
	}
	opts.ll = opts.mdb.DB("dovecot").C("lastlogin")
}

func main() {
	flag.Usage = usage
	flag.Parse()

	if opts.Version {
		fmt.Println(os.Args[0], _VERSION)
		os.Exit(0)
	}

	fs, err := os.OpenFile(opts.LogFile, os.O_WRONLY|os.O_APPEND|os.O_CREATE, 0666)
	if err != nil {
		fmt.Println("Log file error: ", err.Error())
		os.Exit(-4)
	}

	log.SetOutput(fs)
	//	log.SetPrefix("[llmongo] ")

	start := time.Now()
	fmt.Printf("Start: %+v\n", opts)
	log.Printf("Start: %+v\n", opts)

	connectRedis()
	defer opts.rdb.Close()

	connectMongo()
	defer opts.mdb.Close()

	//	// Estrae la lista degli utenti che hanno fatto login negli ultimi X min.
	//	llindex := opts.rdb.Cmd("smembers", "llindex")
	//	lista, err := llindex.List()
	//	if err != nil {
	//		log.Panicln("LLINDEX error: ", err.Error())
	//	}
	//		//	for _, user := range lista {
	//		// cicla fino a che esistono righe di login

	for {
		// estrae un userid dalla lista degli utenti che hanno fatto login
		spop := opts.rdb.Cmd("spop", "llindex")
		user, err := spop.Str()
		log.Printf("SPOP: %+v %+v\n", spop, user)
		// se non ci sono piu' userid esce
		if err != nil {
			log.Printf("LLINDEX empty: %v\n", err)
			break
		}
		//		user := spop.String()
		var date int64
		var lastval, val string
		for {
			// Estrae l'ultimo login dell'utente 'user'
			val, err = opts.rdb.Cmd("lindex", user, "-1").Str()
			if err != nil {
				log.Printf("LINDEX error: %+v\n", err)
				// se ha trovato user e righe di login
				if lastval != "" {
					// reinserisce l'ultimo login e imposta il ttl su Redis
					retval := opts.rdb.Cmd("lpush", user, lastval)
					ttl := opts.rdb.Cmd("expire", user, opts.RedisTTL.Seconds())
					log.Println("LPUSH retval: ", retval, ttl, user, lastval, opts.RedisTTL.Seconds())
				}
				break
			}
			// se la riga di login e' vuota
			if val == "" {
				log.Println("Login empty: ", user)
				retval := opts.rdb.Cmd("lrem", user, "-1", val)
				log.Println("LREM retval: ", user, val, retval)
				continue
			}
			sval := strings.Split(val, ":")
			// se il formato della riga di login non e' corretto
			if sval[1] == "" {
				log.Println("Login format error: ", val, user)
				retval := opts.rdb.Cmd("lrem", user, "-1", val)
				log.Println("LREM retval: ", user, val, retval)
				continue
			}
			// se il timestamp della riga di login non e' corretto
			date, err = strconv.ParseInt(sval[1], 10, 64)
			if err != nil {
				log.Printf("Date Error: %+v - %s\n", err, user)
				continue
			}
			ml := MongoLogin{
				User:     user,
				Protocol: sval[0],
				Ip:       sval[2],
				Date:     time.Unix(date, 0),
			}
			ind := Index{
				User: user,
				Date: time.Unix(date, 0),
			}
			// inserisce il login su Mongodb
			_, err = opts.ll.Upsert(ind, ml)
			if err != nil {
				log.Printf("Insert error: %+v\n", err)
				// se l'errore non e' "duplicate key error" salta al prossimo senza cancellare niente
				if !strings.Contains(err.Error(), "E11000") {
					continue
				}
			}
			// cancella da Redis la riga di login inserita
			retval := opts.rdb.Cmd("lrem", user, "-1", val)
			log.Println("LREM retval: ", retval, user, val)
			lastval = val
		}
		// controlla se ci sono ancora line di login per l'utente 'user'
		/*
			llen, _ := opts.rdb.Cmd("llen", user).Int64()
			if llen <= 1 {
				// elimina l'utente dalla lista di quelli che hanno fatto login se ci sono 1 o meno
				retval := opts.rdb.Cmd("srem", "llindex", user)
				log.Println("SREM retval: ", retval, user)
			}
		*/
	}

	fmt.Printf("Stop %v\n", time.Since(start))
	log.Printf("Stop %v\n", time.Since(start))
}
